@@ -1,27 +1,24 @@
+import asyncio
 import inspect
 import json
 import re
+from collections import Counter
 from collections.abc import AsyncGenerator
+from collections.abc import Awaitable
+from collections.abc import Generator
+from collections.abc import Iterable
+from collections.abc import Iterator
+from collections.abc import Mapping
+from collections.abc import Sequence
 from functools import cached_property
 from functools import partial
 from inspect import Signature
 from re import Pattern
+from typing import Annotated
 from typing import Any
-from typing import Awaitable
 from typing import Callable
 from typing import ClassVar
-from typing import Counter
-from typing import Dict
-from typing import Generator
-from typing import Iterable
-from typing import Iterator
-from typing import List
-from typing import Mapping
 from typing import Optional
-from typing import Sequence
-from typing import Set
-from typing import Tuple
-from typing import Type
 from typing import TypeVar
 from typing import Union
 
@@ -29,6 +26,9 @@ from amgi_types import AMGIReceiveCallable
 from amgi_types import AMGISendCallable
 from amgi_types import LifespanShutdownCompleteEvent
 from amgi_types import LifespanStartupCompleteEvent
+from amgi_types import MessageAckEvent
+from amgi_types import MessageNackEvent
+from amgi_types import MessageReceiveEvent
 from amgi_types import MessageScope
 from amgi_types import MessageSendEvent
 from amgi_types import Scope
@@ -40,7 +40,6 @@ from pydantic.json_schema import GenerateJsonSchema
 from pydantic.json_schema import JsonSchemaMode
 from pydantic.json_schema import JsonSchemaValue
 from pydantic_core import CoreSchema
-from typing_extensions import Annotated
 from typing_extensions import get_args
 from typing_extensions import get_origin
 
@@ -55,9 +54,9 @@ _PARAMETER_PATTERN = re.compile(r"{(.*)}")
 class Message(Mapping[str, Any]):
 
     __address__: ClassVar[Optional[str]] = None
-    __headers__: ClassVar[Dict[str, TypeAdapter[Any]]]
-    __parameters__: ClassVar[Dict[str, TypeAdapter[Any]]]
-    __payload__: ClassVar[Optional[Tuple[str, TypeAdapter[Any]]]]
+    __headers__: ClassVar[dict[str, TypeAdapter[Any]]]
+    __parameters__: ClassVar[dict[str, TypeAdapter[Any]]]
+    __payload__: ClassVar[Optional[tuple[str, TypeAdapter[Any]]]]
 
     def __init_subclass__(cls, address: Optional[str] = None, **kwargs: Any) -> None:
         cls.__address__ = address
@@ -114,7 +113,7 @@ class Message(Mapping[str, Any]):
 
         return self.__address__.format(**parameters)
 
-    def _get_headers(self) -> Iterable[Tuple[bytes, bytes]]:
+    def _get_headers(self) -> Iterable[tuple[bytes, bytes]]:
         return [
             (name.encode(), self._get_value(name, type_adapter))
             for name, type_adapter in self.__headers__.items()
@@ -137,7 +136,7 @@ class Message(Mapping[str, Any]):
 def _generate_message_annotations(
     address: Optional[str],
     fields: dict[str, Any],
-) -> Generator[Tuple[str, Type[Annotated[Any, Any]]], None, None]:
+) -> Generator[tuple[str, type[Annotated[Any, Any]]], None, None]:
     address_parameters = _get_address_parameters(address)
 
     for name, field in fields.items():
@@ -149,11 +148,18 @@ def _generate_message_annotations(
             yield name, Annotated[field, Payload()]  # type: ignore[misc]
 
 
+def _is_message(cls: type[Any]) -> bool:
+    try:
+        return issubclass(cls, Message)
+    except TypeError:
+        return False
+
+
 class AsyncFast:
     def __init__(
         self, title: Optional[str] = None, version: Optional[str] = None
     ) -> None:
-        self._channels: List[Channel] = []
+        self._channels: list[Channel] = []
         self._title = title or "AsyncFast"
         self._version = version or "0.1.0"
 
@@ -175,20 +181,16 @@ class AsyncFast:
 
         messages = []
         return_annotation = signature.return_annotation
-        if (
-            return_annotation is not Signature.empty
-            and get_origin(return_annotation) is AsyncGenerator
+        if return_annotation is not Signature.empty and (
+            get_origin(return_annotation) is AsyncGenerator
+            or get_origin(return_annotation) is Generator
         ):
             async_generator_type = get_args(return_annotation)[0]
             if get_origin(async_generator_type) is Union:  # type: ignore[comparison-overlap]
                 messages = [
-                    type
-                    for type in get_args(async_generator_type)
-                    if inspect.isclass(type) and issubclass(type, Message)
+                    type for type in get_args(async_generator_type) if _is_message(type)
                 ]
-            elif inspect.isclass(async_generator_type) and issubclass(
-                async_generator_type, Message
-            ):
+            elif _is_message(async_generator_type):
                 messages = [get_args(return_annotation)[0]]
 
         annotations = list(_generate_annotations(address, signature))
@@ -249,7 +251,7 @@ class AsyncFast:
                     await channel(scope, receive, send, parameters)
                     break
 
-    def asyncapi(self) -> Dict[str, Any]:
+    def asyncapi(self) -> dict[str, Any]:
         schema_generator = GenerateJsonSchema(
             ref_template="#/components/schemas/{model}"
         )
@@ -273,7 +275,7 @@ class AsyncFast:
 
     def _generate_inputs(
         self,
-    ) -> Generator[Tuple[int, JsonSchemaMode, CoreSchema], None, None]:
+    ) -> Generator[tuple[int, JsonSchemaMode, CoreSchema], None, None]:
         for channel in self._channels:
             headers_model = channel.headers_model
             if headers_model:
@@ -299,7 +301,7 @@ class AsyncFast:
 def _generate_annotations(
     address: str,
     signature: Signature,
-) -> Generator[Tuple[str, Type[Annotated[Any, Any]]], None, None]:
+) -> Generator[tuple[str, type[Annotated[Any, Any]]], None, None]:
 
     address_parameters = _get_address_parameters(address)
 
@@ -316,6 +318,70 @@ def _generate_annotations(
             yield name, Annotated[annotation, Payload()]  # type: ignore[misc]
 
 
+async def _send_message(
+    send_message: Mapping[str, Any], send: AMGISendCallable
+) -> None:
+    message_send_event: MessageSendEvent = {
+        "type": "message.send",
+        "address": send_message["address"],
+        "headers": send_message["headers"],
+        "payload": send_message.get("payload"),
+    }
+    await send(message_send_event)
+
+
+async def _handle_async_generator(
+    handler: Callable[..., AsyncGenerator[Any, None]],
+    arguments: dict[str, Any],
+    send: AMGISendCallable,
+) -> None:
+    agen = handler(**arguments)
+    exception: Optional[Exception] = None
+    while True:
+        try:
+            if exception is None:
+                send_message = await agen.__anext__()
+            else:
+                send_message = await agen.athrow(exception)
+            try:
+                await _send_message(send_message, send)
+            except Exception as e:
+                exception = e
+            else:
+                exception = None
+        except StopAsyncIteration:
+            break
+
+
+def _throw_or_none(gen: Generator[Any, None, None], exception: Exception) -> Any:
+    try:
+        return gen.throw(exception)
+    except StopIteration:
+        return None
+
+
+async def _handle_generator(
+    handler: Callable[..., Generator[Any, None, None]],
+    arguments: dict[str, Any],
+    send: AMGISendCallable,
+) -> None:
+    gen = handler(**arguments)
+    exception: Optional[Exception] = None
+    while True:
+        if exception is None:
+            send_message = await asyncio.to_thread(next, gen, None)
+        else:
+            send_message = await asyncio.to_thread(_throw_or_none, gen, exception)
+        if send_message is None:
+            break
+        try:
+            await _send_message(send_message, send)
+        except Exception as e:
+            exception = e
+        else:
+            exception = None
+
+
 class Channel:
 
     def __init__(
@@ -325,8 +391,8 @@ class Channel:
         handler: Callable[..., Awaitable[None]],
         headers: Mapping[str, TypeAdapter[Any]],
         parameters: Mapping[str, TypeAdapter[Any]],
-        payload: Optional[Tuple[str, TypeAdapter[Any]]],
-        messages: Sequence[Type[Message]],
+        payload: Optional[tuple[str, TypeAdapter[Any]]],
+        messages: Sequence[type[Message]],
     ) -> None:
         self._address = address
         self._address_pattern = address_pattern
@@ -353,7 +419,7 @@ class Channel:
         return self._headers
 
     @cached_property
-    def headers_model(self) -> Optional[Type[BaseModel]]:
+    def headers_model(self) -> Optional[type[BaseModel]]:
         if self._headers:
             headers_name = f"{self.title}Headers"
             headers_model = create_model(
@@ -368,7 +434,7 @@ class Channel:
         return None
 
     @property
-    def payload(self) -> Optional[Tuple[str, TypeAdapter[Any]]]:
+    def payload(self) -> Optional[tuple[str, TypeAdapter[Any]]]:
         return self._payload
 
     @property
@@ -376,10 +442,10 @@ class Channel:
         return self._parameters
 
     @property
-    def messages(self) -> Sequence[Type[Message]]:
+    def messages(self) -> Sequence[type[Message]]:
         return self._messages
 
-    def match(self, address: str) -> Optional[Dict[str, str]]:
+    def match(self, address: str) -> Optional[dict[str, str]]:
         match = self._address_pattern.match(address)
         if match:
             return match.groupdict()
@@ -390,27 +456,45 @@ class Channel:
         scope: MessageScope,
         receive: AMGIReceiveCallable,
         send: AMGISendCallable,
-        parameters: Dict[str, str],
+        parameters: dict[str, str],
     ) -> None:
-        arguments = dict(self._generate_arguments(scope, parameters))
-        if inspect.isasyncgenfunction(self._handler):
-            async for message in self._handler(**arguments):
-                message_send_event: MessageSendEvent = {
-                    "type": "message.send",
-                    "address": message["address"],
-                    "headers": message["headers"],
-                    "payload": message.get("payload"),
+        more_messages = True
+        while more_messages:
+            message = await receive()
+            if message["type"] != "message.receive":
+                continue
+            more_messages = message.get("more_messages", False)
+            try:
+                arguments = dict(self._generate_arguments(message, parameters))
+
+                if inspect.isasyncgenfunction(self._handler):
+                    await _handle_async_generator(self._handler, arguments, send)
+                elif inspect.isgeneratorfunction(self._handler):
+                    await _handle_generator(self._handler, arguments, send)
+                elif inspect.iscoroutinefunction(self._handler):
+                    await self._handler(**arguments)
+                else:
+                    await asyncio.to_thread(self._handler, **arguments)
+
+                message_ack_event: MessageAckEvent = {
+                    "type": "message.ack",
+                    "id": message["id"],
                 }
-                await send(message_send_event)
-        else:
-            await self._handler(**arguments)
+                await send(message_ack_event)
+            except Exception as e:
+                message_nack_event: MessageNackEvent = {
+                    "type": "message.nack",
+                    "id": message["id"],
+                    "message": str(e),
+                }
+                await send(message_nack_event)
 
     def _generate_arguments(
-        self, scope: MessageScope, parameters: Dict[str, str]
-    ) -> Generator[Tuple[str, Any], None, None]:
+        self, message_receive_event: MessageReceiveEvent, parameters: dict[str, str]
+    ) -> Generator[tuple[str, Any], None, None]:
 
         if self.headers:
-            headers = Headers(scope["headers"])
+            headers = Headers(message_receive_event["headers"])
             for name, type_adapter in self.headers.items():
                 annotated_args = get_args(type_adapter._type)
                 header_alias = annotated_args[1].alias
@@ -425,7 +509,7 @@ class Channel:
 
         if self.payload:
             name, type_adapter = self.payload
-            payload = scope.get("payload")
+            payload = message_receive_event.get("payload")
             payload_obj = None if payload is None else json.loads(payload)
             value = type_adapter.validate_python(payload_obj, from_attributes=True)
             yield name, value
@@ -438,7 +522,7 @@ class Channel:
 def _generate_messages(
     channels: Iterable[Channel],
     field_mapping: dict[tuple[int, JsonSchemaMode], JsonSchemaValue],
-) -> Generator[Tuple[str, Dict[str, Any]], None, None]:
+) -> Generator[tuple[str, dict[str, Any]], None, None]:
     for channel in channels:
         message = {}
 
@@ -471,7 +555,7 @@ def _generate_messages(
 
 def _generate_channels(
     channels: Iterable[Channel],
-) -> Generator[Tuple[str, Dict[str, Any]], None, None]:
+) -> Generator[tuple[str, dict[str, Any]], None, None]:
     for channel in channels:
         message_name = f"{channel.title}Message"
         channel_definition = {
@@ -506,7 +590,7 @@ def _generate_channels(
 
 def _generate_operations(
     channels: Iterable[Channel],
-) -> Generator[Tuple[str, Dict[str, Any]], None, None]:
+) -> Generator[tuple[str, dict[str, Any]], None, None]:
     for channel in channels:
         yield f"receive{channel.title}", {
             "action": "receive",
@@ -532,7 +616,7 @@ class Parameter(FieldInfo):
     pass
 
 
-def _get_address_parameters(address: Optional[str]) -> Set[str]:
+def _get_address_parameters(address: Optional[str]) -> set[str]:
     if address is None:
         return set()
     parameters = _PARAMETER_PATTERN.findall(address)
@@ -546,7 +630,7 @@ def _get_address_parameters(address: Optional[str]) -> Set[str]:
 
 class Headers(Mapping[str, str]):
 
-    def __init__(self, raw_list: Iterable[Tuple[bytes, bytes]]) -> None:
+    def __init__(self, raw_list: Iterable[tuple[bytes, bytes]]) -> None:
         self.raw_list = list(raw_list)
 
     def __getitem__(self, key: str, /) -> str:
